/**
 * iSES Remote Lab SDK (c) by J.Dvorak, P.Kuriscak, F.Lustig
 *
 * This file is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 *
 * Copy of license along with attribution guidelines is distributed together with this work in LICENSE.txt file.
 * License text is also available online on <http://creativecommons.org/licenses/by-nc-sa/4.0/>.
 */

/**
 * Worker task used by camera widget to decode images.
 */

importScripts('../vendor/jpg.js');

this.onmessage = function(event) {

	// Handle initialization request.

	if (event.data.type == "init") {
		this.jpeg_decoder = new JpegImage();
		if (this.jpeg_decoder != null) {
			this.postMessage({type: "init"});
		}
		return;
	}

	// Decode the content in the required size.

	var task = event.data;
	this.jpeg_decoder.parse(task.image_bytes);

	// We can only decode three component images so
	// report error back if this is something else.

	if (this.jpeg_decoder.components.length != 3) {
		this.postMessage({type: "error"});
		return;
	}

	var used_width = task.target_width;
	var used_height = task.target_height;
	if (! task.allow_upscale) {
		used_width = Math.min(used_width, this.jpeg_decoder.width);
		used_height = Math.min(used_height, this.jpeg_decoder.height);
	}

	var data = this.jpeg_decoder.getData(used_width, used_height);

	// Send the information back to the main script.

	this.postMessage({
		type: "reply",
		width: used_width,
		height: used_height,
		decoded_width: this.jpeg_decoder.width,
		decoded_height: this.jpeg_decoder.height,
		decode_start_time: task.decode_start_time,
		content: data
	});
};
