/**
 * iSES Remote Lab SDK (c) by J.Dvorak, P.Kuriscak, F.Lustig
 *
 * This file is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 *
 * Copy of license along with attribution guidelines is distributed together with this work in LICENSE.txt file.
 * License text is also available online on <http://creativecommons.org/licenses/by-nc-sa/4.0/>.
 */

$(function ($) {

	$.widget("my.VLButton", {

		options: {
			devices : [],
			descriptionDevice : [],
			description : "NO DESCRIPTION",
			onClickEvent : "",
			autoDisable : true,
			refresh: 100,

			// Locale to use instead of the default one from FORMAT_LOCALE
			locale: null,
		},

		_create: function () {
			if (this.element.length != 1) {
				alert("Button error: Please apply widget to exactly one element!");
			}
			else {
				this._setText(VLH.L("LOADING"));
				if (this.options.devices.length>0) {
					for (var i = 0; i < this.options.devices.length;i++) {
						if (this.options.devices[i].value === undefined) {
							alert("Button error: Please specify value for the device "+(i+1));
						}
					}
				}

				var allDevices = this.options.devices.concat(this.options.descriptionDevice);
				if (allDevices.length>0){
					VLH.GetDeviceHandles(allDevices, $.proxy(this._registerAndUpdate,this));
				} else {
					this._registerAndUpdate();
				}

				this._canBeEnabled = true;
			}
		},



		_registerAndUpdate : function() {
			if (!(this.options.description instanceof Array)) {
				this._setText(this.options.description);
			}
			this.element.button().click($.proxy(function () {this._onClick()},this));
			VL.RegisterEventListener("STATE_CHANGE",this._onStateChange,this);
			if (this.options.descriptionDevice.handle !== undefined) {
				VL.Poll(this.options.descriptionDevice.handle, this.options.refresh, $.proxy(this._onPollUpdate, this));
			}
			this._onStateChange();
		},

		_onPollUpdate: function(result){

			var recalcValue = result.Value;
			if (VLH.IsScalable(this.options.descriptionDevice)){
				recalcValue = VLH.ScaleDeviceValue(result.Value,this.options.descriptionDevice);
			}

			if (this.options.description instanceof Array){
				var i = this.options.description.length-1;
				for (; i > 0; i--) {
					if (this.options.description[i].minValue<=recalcValue) break;
				}

				this._setColor(this.options.description[i].textColor);
				this._setBackgroundColor(this.options.description[i].backgroundColor);

				var used_locale = VLH.getLocale(this.options, null);

				if (this.options.description[i].text !=undefined) {
					this._setText(this.options.description[i].text);
				}
				else if (this.options.description[i].format != undefined) {
					this._setText($.formatNumber(recalcValue, {format: this.options.description[i].format, locale: used_locale}));
				} else {
					this._setText("&nbsp;");
				}

				if (this.options.description[i].enabled != undefined) {
					this._canBeEnabled = this.options.description[i].enabled;
				} else {
					this._canBeEnabled = true;
				}

				this._onStateChange();
			}
		},

		_lastText : undefined,

		_setText : function (new_value) {
			if (this._lastText === new_value) {
				return;
			}
			this._lastText = new_value;

			this.element.html(new_value);
		},

		_lastColor : undefined,

		_setColor : function (new_value) {
			if (this._lastColor === new_value) {
				return;
			}
			this._lastColor = new_value;

			if (new_value !== undefined) {
				this.element.css("color", new_value);
			} else {
				this.element.css("color","");
			}
		},

		_lastBackgroundColor : undefined,

		_setBackgroundColor : function (new_value) {
			if (this._lastBackgroundColor === new_value) {
				return;
			}
			this._lastBackgroundColor = new_value;

			if (new_value !== undefined) {
				this.element.css("background-image","none");
				this.element.css("border-width","3px");
				this.element.css("background-color", new_value);
			} else {
				this.element.css("background-image","inherit");
				this.element.css("border-width","inherit");
				this.element.css("background-color","");
			}
		},

		_setEnableState : function(new_value) {
			var is_enabled = (! this.element.button("option", "disabled"));
			if (is_enabled !== new_value) {
				if (new_value) {
					this.element.button("enable");
				}
				else {
					this.element.button("disable");
				}
			}
		},

		_onStateChange : function() {
			if ((VL.STATE == VL.CS.CONNECTED_CONTROL) && (this._canBeEnabled)){
				this._setEnableState(true);
			}
			else if ((VL.STATE == VL.CS.CONNECTED_WAIT) && (!this.options.autoDisable) && (this._canBeEnabled)) {
				this._setEnableState(true);
			}
			else
			{
				this._setEnableState(false);
			}
		},

		_onClick : function() {
			if (this.options.devices.length==1){
				VL.SetDeviceValue(this.options.devices[0].handle,this.options.devices[0].value,function(){});
			}
			VL.InvokeEvent(this.options.onClickEvent);
			VL.InvokeEvent("USER_ACTIVITY");
		}
	});

	$.widget("my.VLExperimentControl", {

		options: {
			devices : [],
			experimentGroupNumber : 0,
			experimentParameter: 0,
			burstExperimentName: "",
			burstExperimentDuration: 10,
			startEvent : "",
			stopEvent: "",
			onStartEvent: "",
			onStopEvent: "",
			onAutoStopEvent: "",
			experimentNumberSlot: "",
			refresh: 100,
			experimentOverwriteCheckbox: true,
			experimentOverwrite: true
		},

		_create: function () {
			if (this.element.length != 1) {
				alert("Experiment control error: Please apply widget to exactly one element!");
			}
			else {
				if (this.options.burstExperimentName=="") {
					this.element.html(VLH.L("LOADING"));
					if (this.options.devices.length <= 0) {
						alert("Experiment control error: Please specify at least one device.");
					} else {
						VLH.GetDeviceHandles(this.options.devices, $.proxy(function () {
							this.element.html("");
							this._experimentNumberLabel = VLH.CreateTag("div", this.element, "experiment-number", "");
							this._experimentNumberLabel.html("-&nbsp;-&nbsp;-");
							this._experimentId = 0;
							if (this.options.experimentOverwriteCheckbox) {
								this._overwriteCheckBox = VLH.CreateCheckbox(this.element, "overwrite", VLH.L("EXPERIMENT_OVERWRITE"), this.options.experimentOverwrite);
							}
							else {
								this._overwriteCheckBox = undefined;
							}

							VL.RegisterEventListener(this.options.startEvent, this._startExperiment, this);
							VL.RegisterEventListener(this.options.stopEvent, this._stopExperiment, this);

						}, this));
					}
				} else {
					this.element.html("");
					this._experimentNumberLabel = VLH.CreateTag("div", this.element, "experiment-number", "");
					this._experimentNumberLabel.html("-&nbsp;-&nbsp;-");
					this._experimentRunning = false;
					VL.RegisterEventListener(this.options.startEvent, this._startExperiment, this);
				}
			}
		},

		_startExperiment : function (){
			if (this.options.burstExperimentName=="") {
				var expId = 0;
				if (this._overwriteCheckBox !== undefined) {
					expId = this._overwriteCheckBox.prop('checked') ? this._experimentId : 0;
				}
				else {
					expId = this.options.experimentOverwrite ? this._experimentId : 0;
				}

				var handles = [];
				for (var i = 0; i < this.options.devices.length; i++) {
					handles.push(this.options.devices[i].handle);
				}

				VL.StartExperiment(expId, this.options.experimentGroupNumber, 50000, this.options.experimentParameter, false, handles, $.proxy(function (result) {
					if (result.Id != -1) {
						this._experimentRunning = true;
						this._experimentId = result.Id;
						this._experimentNumberLabel.html(this._experimentId);
						VL.Storage[this.options.experimentNumberSlot] = this._experimentId;
						if (this.options.onAutoStopEvent != "") {
							this._pollExperimentRunning();
						}
						VL.InvokeEvent(this.options.onStartEvent);
					}
				}, this));
			} else {
				if (this._experimentRunning) { return; }

				this._experimentRunning = true;
				this._experimentNumberLabel.html("....");
				VL.InvokeEvent(this.options.onStartEvent);
				VL.BurstExperiment(this.options.burstExperimentName, this.options.burstExperimentDuration * 1000, $.proxy(function (result) {
					this._experimentRunning = false;
					this._experimentNumberLabel.html(result.Id);
					VL.Storage[this.options.experimentNumberSlot] = result.Id;
					VL.InvokeEvent(this.options.onAutoStopEvent);
				}, this));
			}
		},

		_stopExperiment : function (){
			VL.StopExperiment($.proxy(function(result){
				if (result.Id!=-1){
					this._experimentRunning = false;
					this._experimentId = result.Id;
					this._experimentNumberLabel.html(this._experimentId);
					VL.Storage[this.options.experimentNumberSlot] = this._experimentId;
					VL.InvokeEvent(this.options.onStopEvent);
				}
			},this));
		},

		_pollExperimentRunning : function(){
			VL.ExperimentRunning(this._experimentId, $.proxy(function(result) {
				if (this._experimentRunning){
					if (result.isRunning){
						setTimeout($.proxy(this._pollExperimentRunning, this), this.options.refresh);
					} else {
						this._experimentRunning = false;
						VL.InvokeEvent(this.options.onAutoStopEvent);
					}
				}
			},this))
		}
	});

	$.widget("my.VLOptionSwitcher", {
		options : {
			elementSelector : "",
			widgetName : "",
			slotName : "",
			triggerEvent : "",
			variants : {}
		},

		_create : function(){
			if (this.element.length != 1) {
				alert("Option switcher error : Please apply widget to exactly one element!");
				return;
			}

			if (this.options.slotName == "") {
				alert("Option switcher error : Please specify slotName!");
				return;
			}

			if (this.options.triggerEvent == "") {
				alert("Option switcher error : Please specify triggerEvent!");
				return;
			}

			var elements = $(this.options.elementSelector);
			if (elements.length!=1){
				alert("Option switcher error : elementSelector should return exactly one element!");
				return;
			}

			this._widget = elements.data("my-"+this.options.widgetName);

			if (this._widget == undefined){
				alert("Options switcher error : Requested widget "+this.options.widgetName+" was not found in the element! Please check that widget is defined before options switcher in javascript code.");
				return;
			}

			VL.RegisterEventListener(this.options.triggerEvent,this._switch,this);
			this._switch();
		},

		_switch : function() {
			var variantValue = this.options.variants[VL.Storage[this.options.slotName]];
			if (variantValue == undefined){
				alert("Options switcher error : Variant "+VL.Storage[this.options.slotName]+ " not found!");
				return;
			}

			$.each( variantValue, $.proxy(function( key, value ) {
				this._widget.option(key, value);
			},this));
		}
	});

	$.widget("my.VLSlotValueSelect" ,{
		options: {
			slotName : "",
			onSelectEvent : "",
			items : [],
			defaultValue : ""
		},

		_create : function(){
			if (this.element.length != 1) {
				alert("Slot value select error : Please apply widget to exactly one element!");
				return;
			}

			if (this.options.slotName == "") {
				alert("Slot value select error : Please specify slotName!");
				return;
			}

			if (this.options.items.length <= 0) {
				alert("Slot value select error : Please specify at least one item!");
				return;
			}

			this.element.html("");
			this._dropDown = VLH.CreateTag("select",this.element,"dropdown","");
			for (var i = 0 ; i < this.options.items.length  ; i++) {
				var selected = (this.options.defaultValue == this.options.items[i].value);
				VLH.CreateOption(this._dropDown,this.options.items[i].value,this.options.items[i].name, selected);
			}
			this._onChange();
			this.element.change($.proxy(this._onChange,this));
		},

		_onChange : function(e) {
			VL.Storage[this.options.slotName] = $("option:selected",this._dropDown).val();
			VL.InvokeEvent(this.options.onSelectEvent);
			VL.InvokeEvent("USER_ACTIVITY");
		}
	});

	$.widget("my.VLExperimentSelect", {

		options: {
			experimentNumberSlot : "",
			experimentGroupNumber: 0,
			onSelectEvent : "",
			refreshEvent : ""
		},

		_create: function () {
			this._dropDown = null;
			if (this.element.length != 1) {
				alert("Experiment select error : Please apply widget to exactly one element!");
			}
			else {
				this.element.html(VLH.L("LOADING"));
				VL.ListExperiments(this.options.experimentGroupNumber,0,0, $.proxy(function(result){
					this._fillDropDown(result);
					VL.RegisterEventListener(this.options.refreshEvent,this._refresh,this);
					this.element.change($.proxy(this._onChange,this));
				},this));
			}
		},

		_refresh : function() {
			VL.ListExperiments(this.options.experimentGroupNumber,0,0, $.proxy(function(result){this._fillDropDown(result);},this));
		},

		_onChange : function(e) {
			VL.Storage[this.options.experimentNumberSlot] = $("option:selected",this._dropDown).val();
			VL.InvokeEvent(this.options.onSelectEvent);
			VL.InvokeEvent("USER_ACTIVITY");
		},

		_fillDropDown : function(result){
			var old_selected = null;
			if (VL.Storage[this.options.experimentNumberSlot] != undefined) {
				old_selected = VL.Storage[this.options.experimentNumberSlot];
			}

			this.element.html("");
			this._dropDown = VLH.CreateTag("select",this.element,"dropdown","");
			VLH.CreateOption(this._dropDown,"-1","");
			for (var i = result.Experiments.length-1 ; i >= 0  ; i--) {
				var startString = new Date(result.Experiments[i].StartDateTimeInSeconds*1000).toDateString();
				var lenString = (result.Experiments[i].EndDateTimeInSeconds-result.Experiments[i].StartDateTimeInSeconds) + " s";
				var selected = (old_selected == result.Experiments[i].Id);
				VLH.CreateOption(this._dropDown,result.Experiments[i].Id,result.Experiments[i].Id + " : " + startString + " - " + lenString, selected);
			}
		}
	});

	$.widget("my.VLDatePicker", {
		options: {
			dateSlot: "",
			defaultDate: new Date(),
			dateFormat: "yy-mm-dd",
			endOfDay : false
		},

		_input: undefined,

		_create: function() {
			if(this.options.dateSlot=="") {
				alert("Date picker error: Please specify slot name.");
			}

			var customOnSelect;

			if (this.options.onSelect!=undefined){
				customOnSelect = this.options.onSelect;
			} else {
				customOnSelect = function(){};
			}

			var updateInput = function () {
				VL.InvokeEvent("USER_ACTIVITY");
				var add = 0;
				if (this.options.endOfDay){
					add = 86400;
				}
				VL.Storage[this.options.dateSlot] = this._input.datepicker('getDate').getTime()/1000+add;
				customOnSelect();
			};

			this.options.onSelect = $.proxy(updateInput,this);

			this._input = VLH.CreateTag("input",this.element,"dateinput","");
			this._input.change($.proxy(updateInput,this));
			this._input.datepicker(this.options);
			this._input.datepicker('setDate',this.options.defaultDate);

			var add = 0;
			if (this.options.endOfDay){
				add = 86400;
			}
			VL.Storage[this.options.dateSlot] = this._input.datepicker('getDate').getTime()/1000+add;
		}
	});

	$.widget("my.VLDataOutput", {

		options: {
			devices : [],
			loadEvent : "",
			formatFunc: VLH.formatCSV,
			formatParams : { },
			exportFunc : VLH.exportToFile,
			exportParams : { },
			postProcs : []
		},

		_create: function () {
			if (! ("element" in this.options.exportParams)) {
				this.options.exportParams.element = this.element;
			}
			if (this.element.length != 1) {
				alert("Experiment data error: Please apply widget to exactly one element!");
			}
			else {
				if (this.options.devices.length<=0){
					alert("Experiment data error: Please specify at least one device.");
				} else {
					VL.RegisterEventListener(this.options.loadEvent,this._loadData,this);
					this._state = {};
				}
			}
		},

		_loadData : function() {

			var _outputFunc = this.options.exportFunc();

			if (_outputFunc.pre != null) {
				_outputFunc.pre(this.options.exportParams);
			}

			var data = VLH.InitDataFromDevices(this.options.devices);
			var postProcState = VLH.InitPostProc(this.options.postProcs);

			VLH.LoadData(this.options.devices, data, this._state, $.proxy(function(){
				var processedData = VLH.PostProc(this.options.postProcs,data,postProcState);
				var content = this.options.formatFunc(processedData,this.options.formatParams);
				_outputFunc.post(content,this.options.exportParams);
			},this));
		}
	});

	$.widget("my.VLCamera" , {
		options: {
			width:320,
			height:240,
			address:"",
			dropCount: 0,

			// Uses native support for M-JPEG streams and fast decoding in some browsers. Entirely relies on the
			// browser to handle the streaming so additional features controlled by the options mentioned bellow
			// are not available.
			forceMjpeg: false,

			// Up-scale image to fit the content area if the received image is too small.
			//
			// Can be disabled to possibly improve performance.
			growImage: true,

			// Resize the image in JavaScript instead of relying on the browser to do it.
			//
			// Currently respected only for upscaling, downscaling is always done in JS.
			swResize: false,

			// Maximal frame rate to render with. Excessive images, if sent, will be skipped during rendering. Ideally
			// should match the frame rate resulting from combination of camera setup and dropCount value to avoid
			// wasting on bandwidth.
			maxFrameRate: 15,

			// Minimal time in seconds between end of decoding and display of one image and start of next one. Gives
			// time to process other parts of the page (e.g. the experiment iself) if image decoding is too slow.
			minFreeTime: 0.02,

			// Tries to do decoding in another thread if possible.
			useWebWorkers: true
		},

		_create: function() {
			if (this.element.length != 1) {
				alert("Camera display error: Please apply widget to exactly one element!");
				return;
			}
			if (this.options.address==""){
				alert("Camera display error: Please specify valid camera address!");
				return;
			}

			// Initialize timing support.

			if ((window.performance != undefined) && (window.performance.now != undefined)) {
				this.getTime = this.getTimePerformance;
			}
			else {
				this.getTime = this.getTimeDate;
			}

			// Force use of mjpeg stream if explicitly requested by user or if websockets
			// are not supported or if we detected a browser where this is a best way (TODO)

			var use_mjpeg = this.options.forceMjpeg || (! ("WebSocket" in window));

			if (! use_mjpeg) {
				var canvas_tag = VLH.CreateTag("canvas", this.element, "cameraCanvas", "", "camera");
				canvas_tag.attr("width", this.options.width);
				canvas_tag.attr("height", this.options.height);
				this.canvas = canvas_tag[0];
				this.retry_timeout = 1000;

				this.resizing_canvas = $("<canvas>");
				this.configured_width = 0;
				this.configured_height = 0;
				this.configureForImageSize(this.canvas.width, this.canvas.height);

				var url = 'ws://' + this.options.address;
				if (this.options.dropCount > 0) {
					url += "?drop=" + this.options.dropCount;
				}
				this.connect_ws(url);

				// Start worker.

				this.web_worker_used = false;

				var workers_supported =
					(typeof(Worker) !== "undefined")
				;

				// Determine path of the worker script. The path is interpeted as relative
				// from the page and not from this script so we need to construct the proper
				// one.

				var worker_script_path = "";
				var scripts = $("script[src]");
				for (var i = 0; i < scripts.length; ++i) {
					var src = scripts[i].src;
					if (src.search("virtuallab/virtuallab_components.js") >= 0) {
						worker_script_path = src.replace("virtuallab/virtuallab_components.js", "virtuallab/virtuallab_camera_worker.js");
						break;
					}
				}

				if (this.options.useWebWorkers && workers_supported && (worker_script_path != "")) {

					this.worker_image_pending = 0;
					this.worker_processing_durations = [];

					try {

						this.worker = new Worker(worker_script_path);
						this.worker.addEventListener('message', $.proxy(this.updateImageWorkerFinished, this));

						// If there is any error, we will stop using the worker.

						this.worker.addEventListener('error', $.proxy(function () {
							this.web_worker_used = false;
						}, this));

						// Let the worker do its initialization.

						this.worker.postMessage({type: "init"});
					}
					catch (error) {
						console.log("Web workers init failed: " + error);
					}
				}

				// Start update timer.

				this.imageData = null;
				this.scheduleNextUpdateTimeout(0);
			}
			else {
				this.connect_mjpeg();
			}

		},

		get_mjpeg_url: function() {

			// Add random element to URL to prevent any caching.

			var url = 'http://' + this.options.address + "?now=" + new Date().getTime();
			if (this.options.dropCount > 0) {
				url += "&drop=" + this.options.dropCount;
			}
			return url;
		},

		connect_mjpeg: function() {
			this.element.html("");
			var img_tag = VLH.CreateTag("img", this.element, "cameraImg", "", "camera");

			// On error reload the image. Note that this is triggered only if the load fails entirely (e.g. image
			// server is not running or the connection to it failed). We do not have way to detect if the communication
			// stops latter.

			img_tag.on("error", $.proxy(function() {setTimeout($.proxy(function() {img_tag.attr("src", this.get_mjpeg_url()); }, this), 5000)} , this));
			img_tag.attr("src", this.get_mjpeg_url());
		},

		connect_ws: function(address) {
			this.jpeg_decoder = new JpegImage();
			this.socket = new WebSocket(address);
			this.socket.binaryType = "arraybuffer";

			this.socket.onopen = $.proxy(function() {
				this.image_id = 0;
				this.retry_timeout = 1000;
			},this);

			this.socket.onclose = $.proxy(function() {
				setTimeout($.proxy(function() {this.connect_ws(address);}, this), this.retry_timeout);

				// Increase timeout after each unsuccessful attempt.

				this.retry_timeout = Math.min(this.retry_timeout + 1000, 10000);
			},this);

			this.socket.onmessage = $.proxy(function(evt) {
				++this.image_id;
				//console.log("Received image " + this.image_id);

				// Send confirmation messages.

				var out_buffer = new ArrayBuffer(8);
				var out_view = new DataView(out_buffer);
				out_view.setUint32(0, (this.image_id &  0xffffffff), true);
				out_view.setUint32(4, (this.image_id / 0x100000000), true);
				this.socket.send(out_buffer);

				// Store the data for the update timer to process.

				this.imageData = new Uint8Array(evt.data);
			}, this);
		},

		// Ensures that we are configured for rendering of image of specified dimensions
		//
		// Does nothing if we are already configured.
		//
		configureForImageSize : function(width, height) {
			if ((width == this.configured_width) && (height == this.configured_height)) {
				return;
			}
			this.configured_width = width;
			this.configured_height = height;

			this.final_image_width = this.options.growImage ? this.canvas.width : Math.min(this.canvas.width, width);
			this.final_image_height = this.options.growImage ? this.canvas.height : Math.min(this.canvas.height, height);
			this.final_image_pos_x = (this.canvas.width - this.final_image_width) / 2;
			this.final_image_pos_y = (this.canvas.height - this.final_image_height) / 2;

			// Clear entire area of the canvas. Note that we do this
			// only when the dimensions change which should normally not
			// happen without user interaction.

			this.canvas.getContext("2d").clearRect(0, 0, this.canvas.width, this.canvas.height);

			// Resize the canvas we are using to get HW accelerated resizes.

			this.resizing_canvas.attr("width", width);
			this.resizing_canvas.attr("height", height);
		},

		// Draws the image using the data we received since the last call.
		//
		updateImage : function() {
			if (!this.web_worker_used) {
				this.updateImageNoWorker();
			}
			else {
				this.updateImageWorkerStart();
			}
		},

		// Let the worker do its work.
		//
		updateImageWorkerStart : function() {
			if (this.imageData == null) {
				this.scheduleNextUpdateTimeout(0);
				return;
			}

			// The worker is still busy with something, skip this update.

			if (this.worker_image_pending) {
				//console.log("Worker still busy");
				this.scheduleNextUpdateTimeout(0);
				return;
			}

			// Let the worker do the decoding. We will
			// get message when it is done.

			this.worker_image_pending = true;
			this.worker.postMessage({
				type: "decode",
				target_width: this.final_image_width,
				target_height: this.final_image_height,
				image_bytes: this.imageData,
				decode_start_time: this.getTime(),
				allow_upscale: this.options.growImage && this.options.swResize
			});

			// Drop the old data we do not want to do anything until we get something
			// new.

			this.imageData = null;

			// Schedule next update. While we did end immediately on the main thread, there will
			// be some delay before the processing completes so account for that.

			var max_duration = 0;
			if (this.worker_processing_durations.length > 0) {
				for (var i = 0; i < this.worker_processing_durations.length; ++i) {
					max_duration = Math.max(max_duration, this.worker_processing_durations[i]);
				}
			}
			this.scheduleNextUpdateTimeout(max_duration);
		},

		// Called when the worker ends its task.
		//
		updateImageWorkerFinished : function(event) {

			// Is this a notification that the worker completed its initialization?

			if (event.data.type == "init") {
				this.web_worker_used = true;
				return;
			}
			else if (event.data.type == "error") {
				this.web_worker_used = false;
				return;
			}

			// Ignore messages from the worker once we decide that we are not going
			// to use it. This is permanent state.

			if (! this.web_worker_used) {
				return;
			}

			this.worker_image_pending = false;

			// Update for the real image size we received from the worker.

			this.configureForImageSize(event.data.decoded_width, event.data.decoded_height);

			// If the data have the required size, we will put them directly
			// into the target context. Otherwise they will be resized by browser.

			var output_context = this.canvas.getContext("2d");
			var data_context;
			var do_blit;
			if ((event.data.width != this.final_image_width) || (event.data.height != this.final_image_height)) {
				data_context = this.resizing_canvas[0].getContext("2d");
				do_blit = true;
			}
			else {
				data_context = output_context;
				do_blit = false;
			}

			var image = data_context.createImageData(event.data.width, event.data.height);
			var byte_count = event.data.width * event.data.height * 3;

			var input_array = event.data.content;
			var output_array = image.data;

			var output_index = 0;
			var input_index = 0;

			while (input_index < byte_count) {
				output_array[output_index++] = input_array[input_index++];
				output_array[output_index++] = input_array[input_index++];
				output_array[output_index++] = input_array[input_index++];
				output_array[output_index++] = 255;
			}

			// Put the image to the output.

			if (do_blit) {
				data_context.putImageData(image, 0, 0);
				output_context.drawImage(this.resizing_canvas[0], 0, 0, event.data.width, event.data.height, this.final_image_pos_x, this.final_image_pos_y, this.final_image_width, this.final_image_height);
			}
			else {
				output_context.putImageData(image, this.final_image_pos_x, this.final_image_pos_y);
			}

			// Update our information about how long it did take the worker to
			// decode the image and us to process it so we can properly schedule the
			// next update.

			var processing_duration = this.getTime() - event.data.decode_start_time;
			this.worker_processing_durations.push(processing_duration);
			while (this.worker_processing_durations.length > 10) {
				this.worker_processing_durations.shift();
			}
			//console.log("Decoding finished " + processing_duration);
		},

		// Draws the image without using the worker thread.
		//
		updateImageNoWorker : function() {
			if (this.imageData == null) {
				this.scheduleNextUpdateTimeout(0);
				return;
			}

			// Decode content.

			var processing_start = this.getTime();
			this.jpeg_decoder.parse(this.imageData);

			var decoded_width = this.jpeg_decoder.width;
			var decoded_height = this.jpeg_decoder.height;

			// Ensure that we are ready for this image in case it changed
			// during operation.

			this.configureForImageSize(decoded_width, decoded_height);

			var output_context = this.canvas.getContext("2d");

			// If there no upscale necessary or if we are asked to do it in SW, let the decoder do it and put
			// the data directly to the target canvas.

			var data_context;
			var data_width, data_height;
			var do_blit;
			if (this.options.swResize || ((this.final_image_width <= decoded_width) && (this.final_image_height <= decoded_height))) {
				data_context = output_context;
				data_width = this.final_image_width;
				data_height = this.final_image_height;
				do_blit = false;
			}

			// Otherwise we put it into secondary canvas of the data size.

			else {
				data_context = this.resizing_canvas[0].getContext("2d");
				data_width = decoded_width;
				data_height = decoded_height;
				do_blit = true;
			}

			var data = data_context.createImageData(data_width, data_height);
			this.jpeg_decoder.copyToImageData(data);

			if (do_blit) {
				data_context.putImageData(data, 0, 0);
				output_context.drawImage(this.resizing_canvas[0], 0, 0, data_width, data_height, this.final_image_pos_x, this.final_image_pos_y, this.final_image_width, this.final_image_height);
			}
			else {
				data_context.putImageData(data, this.final_image_pos_x, this.final_image_pos_y);
			}

			var processing_end = this.getTime();
			var processing_duration = processing_end - processing_start;
			//console.log(processing_duration);

			// Drop the old data we do not want to do anything until we get something
			// new.

			this.imageData = null;

			// Schedule when to do the next update.

			this.scheduleNextUpdateTimeout(processing_duration);
		},

		// Schedules next call of _updateImage
		//
		scheduleNextUpdateTimeout : function(processing_duration) {

			// How long we need to wait until next tick on our desired maximal framerate.

			var time_to_next_frame = (1000 / this.options.maxFrameRate) - processing_duration;

			// At least how long we need to wait after current decoding to
			// give required time to other parts of the page.

			var minimal_time_to_next_update = this.options.minFreeTime * 1000;

			var time_to_next_update = Math.max(0, Math.max(minimal_time_to_next_update, time_to_next_frame));
			//console.log("Timeout " + time_to_next_update);
			setTimeout($.proxy(this.updateImage, this), time_to_next_update);
		},

		getTimeDate : function() {
			return Date.now();
		},

		getTimePerformance : function() {
			return performance.now();
		}
	});


	$.widget("my.VLChart" , {
		options: {
			devices : [],
			loadEvents : [],
			loadOnCreate : false,
			loadInterval : -1,
			postProcs : [],
			width: 500,
			height: 300,

			autoScroll: false,
			autoScrollSpeedX : 1,
			autoScrollSpeedY : 0,
			autoScrollFps : 20,
			autoScrollWindowX : 1,
			autoScrollWindowY : 1,

			chartSeries : [],
			chartParams : {},

			// Should the graph show error message (true) or empty content (false)
			// when no data are available.
			requireData: true,

			// Should the graph stop updates and clear itself when the control
			// is lost? When the control is regained, the graph will restart
			// itself depending on value of loadOnCreate parameter
			requireControl: false
		},

		_create: function() {
			this._backup = VLH.backup(this);
			this._activeNotifications = [];
			this._chartIsDisabled = false;

			if (this.element.length != 1) {
				alert("Experiment data error: Please apply widget to exactly one element!");
			}
			else {
				if (this.options.devices.length<=0){
					alert("Experiment data error: Please specify at least one device.");
				} else {

					this._notifications = VLH.CreateTag('div',this.element,'notification','','chart-notification');
					this._chartDiv = VLH.CreateTag('div',this.element,'chart',' ','');

					if (this.element.css('width') == "") {
						this._chartDiv.css('width',this.options.width);
					} else {
						this._chartDiv.css('width','100%');
					}

					if (this.element.css('height') == "") {
						this._chartDiv.css('height',this.options.height);
					} else {
						this._chartDiv.css('height','100%');
					}

					// Prepare series we will use if the default series are not configured. Keeping this as separate
					// object allows us to immediately react to change in the series option without need to reset
					// the graph.

					this._defaultCharSeries = new Array();
					for (var i = 0; i < this.options.devices.length ; i++){
						var line = {};
						line.y = this.options.devices[i].deviceName;
						line.x = this.options.devices[i].deviceName+".time";
						this._defaultCharSeries.push(line);
					}

					this._state = null;
					this._lastChartData = null;
					this._dataStats = {};
					this._loadEventsRegInfo = VL.RegisterEventListener(this.options.loadEvents,this._loadData,this);

					if (this.options.loadInterval>0){
						this._loadIntervalID = setInterval($.proxy(this._loadData,this),this.options.loadInterval*1000);
					}

					if (this.options.autoScroll){
						this._lastUpdateTime = new Date().getTime();
						this._sPosX = 0;
						this._sPosY = 0;
						this._sVelX = this.options.autoScrollSpeedX;
						this._sVelY = this.options.autoScrollSpeedY;
						this._sOffsetX = 0;
						this._sOffsetY = 0;
						this._sOffsetXArray = [];
						this._sOffsetYArray = [];
						this._autoScrollIntervalID = setInterval($.proxy(this._autoScrollDraw,this),1000.0/this.options.autoScrollFps);
					}

					if (this.options.loadOnCreate){
						this._loadData();
					}
					if (this.options.requireControl) {
						VL.RegisterEventListener("STATE_CHANGE",this._onStateChange,this);
					}
				}
			}
		},

		/**
		 * Starts definition of active notifications.
		 **/
		_startNotifications : function() {
			this._knownNotifications = {};
			this._nextNotificationIdx = 0;
			this._notificationsChanged = false;
		},
		
		/**
		 * Adds notification to the active definition.
		 */
		_addNotification : function (message) {

			// Filter out duplicate messages.

			if (message in this._knownNotifications) {
				return;
			}
			this._knownNotifications[message] = true;

			// Look for change at the same slot.

			if (this._nextNotificationIdx < this._activeNotifications.length) {
				if (this._activeNotifications[this._nextNotificationIdx] != message) {
					this._notificationsChanged = true;
					this._activeNotifications[this._nextNotificationIdx] = message;
				}
			}

			// Or extension of the array.

			else {
				this._notificationsChanged = true;
				this._activeNotifications.push(message);
			}
			++this._nextNotificationIdx;
		},

		/**
		 * Adds all notifications from specified array and adds to them specified prefix
		 */
		_addNotifications : function (prefix, messages) {
			for (var i = 0; i < messages.length; ++i) {
				this._addNotification(prefix + messages[i]);
			}
		},

		/**
		 * Ends definition of nactive notifications and updates
		 * the page if there was any change since previous state.
		 **/
		_finishNotifications : function() {

			// Trim whatever additional notifications were there in the
			// previous state.

			if (this._nextNotificationIdx != this._activeNotifications.length) {
				this._notificationsChanged = true;
				this._activeNotifications = this._activeNotifications.slice(0, this._nextNotificationIdx);
			}

			// Update the page if there was some change.

			if (! this._notificationsChanged) {
				return;
			}
			this._notifications.html('');
			for (var i = 0; i < this._nextNotificationIdx; ++i) {
				VLH.CreateTag('div', this._notifications,'ntf', this._activeNotifications[i], 'chart-notification-message');
			}
		},

		/**
		 * Removes previous notification messages.
		 */
		_resetNotifications : function() {
			this._startNotifications();
			this._finishNotifications();
		},

		_showDisabledChart : function(message) {
			if (this._chartIsDisabled) {
				return;
			}
			this._chartDiv.addClass('disabled-chart');

			// The top padding percentage is calculated from parent WIDTH so
			// we need to calculate the percentage which would correspond to
			// 50% of height. 

			var top_padding_percentage = 50 * (this._chartDiv.parent().height() / this._chartDiv.parent().width());
			this._chartDiv.css("padding-top", top_padding_percentage.toString() + "%");
			this._chartDiv.html(VLH.L(message));
			this._chartIsDisabled = true;
			this._lastChartData = null;
		},

		_hideDisabledChart : function() {
			if (! this._chartIsDisabled) {
				return;
			}
			this._chartDiv.removeClass('disabled-chart');
			this._chartDiv.css("padding-top", "");
			this._chartDiv.html(' ');
			this._chartIsDisabled = false;
			this._lastChartData = null;
		},

		_destroy : function() {
			VLH.undefine(this,"_loadIntervalID",clearInterval);
			VLH.undefine(this,"_autoScrollIntervalID",clearInterval);
			VLH.undefine(this,"_loadEventsRegInfo",VL.ClearEventListener);
			VLH.undefine(this,"_state",VLH.StopLoadingData);
			this.element.html("");
			this.element.style = this._initialStyles;
		},

		_setOptions : function(options){
			var reset = false;
			//2DO : Implement different handling of component reset.
			$.each( options, $.proxy(function( key, value ) {
				this._setOption( key, value );
				if ((key == "devices") ||
				   ((key == "loadInterval") && (value != this.options.loadInterval)) ||
					(key == "autoScrollFps") && (value != this.options.autoScrollFps) ||
					(key == "autoScroll")||
					(key == "postProcs") ||
					(key == "loadEvents") ||
					(key == "requireControl")
				) {
					reset = true;
				}
			},this));

			if (reset) {
				this._destroy();
				this._create();
			}
			this._lastChartData = null;
		},

		_onStateChange : function() {
			var has_control = (VL.STATE == VL.CS.CONNECTED_CONTROL);
			if (! has_control) {
				if (this._state != null) {
					VLH.StopLoadingData(this._state);
					this._state = null;
				}
				this._showDisabledChart("CHART_NO_CONTROL");
			}
			else if (this._state == null) {
				if (this.options.loadOnCreate) {
					this._loadData();
				}
				else {
					this._hideDisabledChart();
					this._resetNotifications();

					var dummy_data = this._constructChartData(null, null);
					Flotr.draw(this._chartDiv[0], dummy_data, this.options.chartParams);
				}
			}
		},

		_loadData : function() {

			if (this.options.requireControl && (VL.STATE != VL.CS.CONNECTED_CONTROL)) {
				this._onStateChange();
				return;
			}

			if (this._state != null) {
				VLH.StopLoadingData(this._state);
			}
			this._state = {};
			this._lastChartData = null;
			this._hideDisabledChart();
			this._resetNotifications();
			var data = VLH.InitDataFromDevices(this.options.devices);
			var postProcState = VLH.InitPostProc(this.options.postProcs);
			VLH.LoadData(this.options.devices, data, this._state, $.proxy(function(){
				var processedData = VLH.PostProc(this.options.postProcs,data,postProcState);
				this._chartData = this._constructChartData(processedData,this._dataStats);

				if (!this.options.autoScroll){
					if (this._chartData!=undefined) {
						this._hideDisabledChart();
						if (! this._isEqualChartDataSeries(this._lastChartData, this._chartData)) {
							this._lastChartData = this._chartData;
							Flotr.draw(this._chartDiv[0], this._chartData, this.options.chartParams);
						}
					} else {
						this._showDisabledChart("CHART_ERROR");
					}
				} else {
					this._updatePos();
					if (this._dataStats.hasData){
						var optimalX = this.options.autoScrollSpeedX > 0 ? this._dataStats.maxX - this.options.autoScrollWindowX : this._dataStats.minX;
						var optimalY = this.options.autoScrollSpeedY > 0 ? this._dataStats.maxY - this.options.autoScrollWindowY : this._dataStats.minY;

						if(this.options.autoScrollSpeedX!=0){
							if(Math.abs(this._sPosX-optimalX)>(0.5*this.options.autoScrollWindowX)){
								this._sPosX = optimalX;
								this._sOffsetX = 0;
								this._sOffsetXArray = [];
							} else {
								this._sVelX += (optimalX - this._sPosX)*0.06 - 0.06*(this._sVelX-this.options.autoScrollSpeedX);
								this._sOffsetX = VLH.UpdateAverage((optimalX - this._sPosX),this._sOffsetXArray,20);
							}
						}

						if(this.options.autoScrollSpeedY!=0){
							if(Math.abs(this._sPosY-optimalY)>(0.5*this.options.autoScrollWindowY)){
								this._sPosY = optimalY;
								this._sOffsetY = 0;
								this._sOffsetArray = [];
							} else {
								this._sVelY += (optimalY - this._sPosY)*0.06 - 0.06*(this._sVelY-this.options.autoScrollSpeedY);
								this._sOffsetY = VLH.UpdateAverage((optimalY - this._sPosY),this._sOffsetYArray,20);
							}
						}
					}
				}
			},this));
		},

		_autoScrollDraw : function() {
			this._updatePos();
			var chartParams = { xaxis: {}, yaxis: {}};
			if (this.options.autoScrollSpeedX!=0){
				chartParams.xaxis.min = this._sPosX + this._sOffsetX;
				chartParams.xaxis.max = chartParams.xaxis.min + this.options.autoScrollWindowX;
			}

			if (this.options.autoScrollSpeedY!=0){
				chartParams.yaxis.min = this._sPosY + this._sOffsetY;
				chartParams.yaxis.max = chartParams.yaxis.min + this.options.autoScrollWindowY;
			}
			if (this._chartData!=undefined) {
				this._hideDisabledChart();
				Flotr.draw(this._chartDiv[0], this._chartData, $.extend(true, {}, this.options.chartParams, chartParams));
			} else {
				this._showDisabledChart("CHART_ERROR");
			}
		},

		_updatePos : function() {
			var now = new Date().getTime();
			var dt = (now - this._lastUpdateTime)/1000.0;
			this._lastUpdateTime = now;
			this._sPosX += this._sVelX*dt;
			this._sPosY += this._sVelY*dt;
		},

		// Determines if values of all series in the data are equal
		//
		_isEqualChartDataSeries : function(first, second)
		{
			// Consider two null data to be equal.

			if ((first === null) || (second === null)) {
				return first === second;
			}

			// Same number of series.

			if (first.length !== second.length) {
				return false;
			}

			// Matching series must have the same number of values

			var seriesCount = first.length;
			for (var seriesIdx = 0; seriesIdx < seriesCount; ++seriesIdx) {
				if (first[seriesIdx].data.length !== second[seriesIdx].data.length) {
					return false;
				}
			}

			// We need to compare data in all the series.

			for (var seriesIdx = 0; seriesIdx < seriesCount; ++seriesIdx) {
				var firstSeriesData = first[seriesIdx].data;
				var secondSeriesData = second[seriesIdx].data;
				var valueCount = firstSeriesData.length;

				for (var valueIdx = 0; valueIdx < valueCount; ++valueIdx) {
					if (firstSeriesData[valueIdx][0] !== secondSeriesData[valueIdx][0]) {
						return false;
					}
					if (firstSeriesData[valueIdx][1] !== secondSeriesData[valueIdx][1]) {
						return false;
					}
				}
			}

			// Look the same.

			return true;
		},

		// Prepares data passed to the flotr to draw the grap.
		//
		// processedData can be null in which case the function will behave
		// as if input provided all required series without any values nor
		// errors in log
		_constructChartData : function(processedData,dataStats) {
			var chartData = [];

			if (dataStats!=null){
				var minX = Number.MAX_VALUE;
				var maxX = Number.MIN_VALUE;
				var minY = Number.MAX_VALUE;
				var maxY = Number.MIN_VALUE;
				var hasData = false;
			}
			
			this._startNotifications();

			var exists_valid_chart = false;

			var series = (this.options.chartSeries.length > 0) ? this.options.chartSeries : this._defaultCharSeries;

			for (var i = 0; i < series.length;i++){

				var Xseries;
				var Yseries;
				var colorSeries = null;
				if (processedData != null) {
					Xseries = processedData.series[series[i].x];
					Yseries = processedData.series[series[i].y];
					if ("colors" in series[i]) {
						colorSeries = VLH.getSeries(processedData, series[i].colors);
					}
				}
				else {
					Yseries = Xseries = {log: [], values: []};
				}

				var seriesData = {};
				seriesData.data = [];
				$.extend(seriesData,series[i]);
				chartData.push(seriesData);

				if ((Xseries.log.length>0) || (Yseries.log.length>0)) {
					var name = (series[i].label !== undefined) ? series[i].name : VLH.L('SERIES') + " " + (i + 1);
					var prefix = name + ": ";
					this._addNotifications(prefix, Xseries.log);
					this._addNotifications(prefix, Yseries.log);
					continue;
				}

				if (this.options.requireData && ((Xseries.values.length == 0) || (Yseries.values.length == 0))) {
					var name = (series[i].label !== undefined) ? series[i].name : VLH.L('SERIES') + " " + (i + 1);
					this._addNotification(name + ": " + VLH.L("NO_DATA"));
					continue;
				}

				var numPoints = Math.min(Xseries.values.length,Yseries.values.length);
				var points = seriesData.data;
				for (var j = 0; j < numPoints ; j++){
					if (colorSeries == null) {
						points.push([Xseries.values[j],Yseries.values[j]]);
					}
					else {
						points.push([Xseries.values[j],Yseries.values[j],colorSeries.values[j]]);
					}
					if (dataStats != null){
						hasData = true;
						minX = Math.min(minX,Xseries.values[j]);
						minY = Math.min(minY,Yseries.values[j]);
						maxX = Math.max(maxX,Xseries.values[j]);
						maxY = Math.max(maxY,Yseries.values[j]);
					}
				}
				exists_valid_chart = true;
			}

			if (dataStats!=null){
				dataStats.minX = minX;
				dataStats.maxX = maxX;
				dataStats.minY = minY;
				dataStats.maxY = maxY;
				dataStats.hasData = hasData;
			}

			this._finishNotifications();

			if (exists_valid_chart) {
				return chartData;
			} else {
				return undefined;
			}
		}
	});

	$.widget("my.VLPeriodicSend" , {

		options : {
			devices : [],
			refresh : 1000
		},

		_create: function () {
			if (this.element.length != 1) {
				alert("Periodic send error: Please apply widget to exactly one element!");
			}
			else {
				this.element.html(VLH.L("LOADING"));
				if (this.options.devices.length==0){
					alert("Periodic send error: Please specify at least one device.");
				} else {
					this.element.hide();
					var cthis = this;
					VLH.GetDeviceHandles(this.options.devices,function(){
						cthis._lastSend = new Date().getTime();
						cthis.element.html("OK");
						setTimeout(function () {cthis._send(cthis);},cthis.options.refresh);
					});
				}
			}
		},

		_send : function(cthis) {
			cthis._lastSend = new Date().getTime();
			VLH.ForEachDevice(cthis.options.devices,function(device,deviceCallback) {
				if (device.value!=null){
					VL.SetDeviceValue(device.handle,device.value,deviceCallback);
				} else {
					alert("Periodic send error: Value for device "+ device.deviceName +" not set.");
				}
			}, function () {},
			   function (devices) {
				var milisToGo = (new Date().getTime()) - cthis._lastSend - cthis.options.refresh;
				if (milisToGo > 0) {
					cthis._send(cthis);
				}
				else {
					setTimeout(function () {
						cthis._send(cthis);
					}, -milisToGo);
				}
			});
		}

	});

	$.widget("my.VLValueDisplay" , {
		options : {
			devices : [],
			refresh : 100,
			format : "#.##",

			// Locale to use instead of the default one from FORMAT_LOCALE
			locale: null,

			// Content to show when the user does not have control. Has the same
			// format as the ordinary format parameter. If null, the content is
			// the same as with normal format
			noControlFormat: null
		},

		_create: function () {
			if (this.element.length != 1) {
				alert("Value display error: Please apply widget to exactly one element!");
			} else {
				this._setText(VLH.L("LOADING"));
				this._lastValue = undefined;
				if (this.options.devices.length!=1){
					alert("Value display error: Please specify exactly one device.");
				} else {
					VLH.GetDeviceHandles(this.options.devices, $.proxy(function(){
						VL.RegisterEventListener("STATE_CHANGE",this._updateDisplay,this);
						VL.Poll(this.options.devices[0].handle,this.options.refresh, $.proxy(function(result) {
							var recalcValue = result.Value;
							if (VLH.IsScalable(this.options.devices[0])){
								recalcValue = VLH.ScaleDeviceValue(result.Value,this.options.devices[0]);
							}
							this._lastValue = recalcValue;
							this._updateDisplay();
						},this));
					},this));
				}
			}
		},

		_lastText : undefined,

		_setText : function (new_value) {
			if (this._lastText === new_value) {
				return;
			}
			this._lastText = new_value;

			this.element.html(new_value);
		},

		_lastColor : undefined,

		_setColor : function (new_value) {
			if (this._lastColor === new_value) {
				return;
			}
			this._lastColor = new_value;

			if (new_value !== undefined) {
				this.element.css("color", new_value);
			} else {
				this.element.css("color","");
			}
		},

		_lastBackgroundColor : undefined,

		_setBackgroundColor : function (new_value) {
			if (this._lastBackgroundColor === new_value) {
				return;
			}
			this._lastBackgroundColor = new_value;

			if (new_value !== undefined) {
				this.element.css("background-color", new_value);
			} else {
				this.element.css("background-color","");
			}
		},

		_updateDisplay : function() {

			// Are we still loading the first value?

			if (this._lastValue === undefined) {
				return;
			}

			// Determine set of formating rules to use given the current state.

			var used_format =
				((VL.STATE == VL.CS.CONNECTED_CONTROL) || (this.options.noControlFormat == null)) ?
				this.options.format :
				this.options.noControlFormat
			;

			var used_locale = VLH.getLocale(this.options, null);

			// Apply it.

			if (used_format instanceof Array){
				var i = used_format.length-1;
				for (; i > 0; i--) {
					if (used_format[i].minValue<=this._lastValue) break;
				}

				this._setColor(used_format[i].textColor);
				this._setBackgroundColor(used_format[i].backgroundColor);

				if (used_format[i].text !=undefined) {
					this._setText(used_format[i].text);
				} else {
					this._setText($.formatNumber(this._lastValue, {format: used_format[i].format, locale: used_locale}));
				}
			} else {
				this._setText($.formatNumber(this._lastValue, {format: used_format, locale: used_locale}));
			}
		}
	});

	$.widget("my.VLSlider" , {
		options : {
			devices : [],
			min: 0,
			max: 1,
			select_step : 0.01,
			button_step : 0.1,
			isControlled : true,
			autorepeatInterval: 0.3,
			autorepeatDelay : 3.0,
			orientation: "horizontal"
		},

		_create: function(){
			if (this.element.length != 1) {
				alert("Slider error: Please apply widget to exactly one element!");
			} else {
				this._sliderChanged = false;

				if (this.options.devices.length<=2)
				{
					this._canSend = true;
					this._canPoll = true;

					this._minusButtonElement = VLH.CreateTag("div",this.element,"minus-button");
					this._sliderElement = VLH.CreateTag("div",this.element,"slider");
					this._plusButtonElement = VLH.CreateTag("div",this.element,"plus-button");

					this._sliderElement.slider({
						min: this.options.min,
						max: this.options.max,
						orientation : this.options.orientation,
						step : this.options.select_step,
						value : this.options.min,
						enable : false,

						slide : $.proxy(function(event,ui){
							this._sliderChanged = true;
							this._handleSend(ui.value);
						},this),

						start : $.proxy(function(){this._canPoll = false;},this),
						stop : $.proxy(function(){this._canPoll = true;},this)
					});

					this._minusButtonElement.button({icons : { primary : "ui-icon-triangle-1-w" }})
						.click($.proxy(this._stepLeft, this))
						.mousedown($.proxy(function() {this._startPressRepeat(this._stepLeft);}, this))
						.mouseup($.proxy(this._stopPressRepeat, this))
						.mouseleave($.proxy(this._stopPressRepeat, this))
						.css('width','35px').css('float','left');

					this._plusButtonElement.button({icons : { primary : "ui-icon-triangle-1-e" }})
						.click($.proxy(this._stepRight, this))
						.mousedown($.proxy(function() {this._startPressRepeat(this._stepRight);}, this))
						.mouseup($.proxy(this._stopPressRepeat, this))
						.mouseleave($.proxy(this._stopPressRepeat, this))
						.css('width','35px').css('float','right');

					this._sliderElement.css('float','left').width(this.element.width()-81);

					this._changeState("disable");

					VLH.GetDeviceHandles(this.options.devices, $.proxy(function(){
						VL.RegisterEventListener("STATE_CHANGE",this._onStateChange,this);
						this._onStateChange();
						if (this.options.devices.length == 2){
							_outIndex = 0;
							_inIndex = 1;
						} else {
							_outIndex = 0;
							_inIndex = 0;
						}
						if (this.options.isControlled){
							VL.Poll(this.options.devices[_inIndex].handle,this.options.devices[_inIndex].refresh, $.proxy(function(result){
								if ((this._canSend) && (this._canPoll) && (this._repeatTimer === undefined)){
									this._sliderElement.slider("value",result.Value);
								}
							},this));
						}
					},this));

				} else {
					alert("Slider error : Please specify exactly one device.");
				}
			}
		},

		_destroy : function() {
			this._stopPressRepeat();
		},

		_stepLeft : function() {
			this._stepSlider(-1);
		},

		_stepRight : function() {
			this._stepSlider(1);
		},

		_stepSlider : function(direction) {
			this._sliderChanged = true;
			var new_value = VLH.Clamp(this._sliderElement.slider("value") + (direction * this.options.button_step),this.options.min,this.options.max);
			this._sliderElement.slider("value",new_value);
			this._handleSend(new_value);
		},

		_startPressRepeat : function(callback) {
			this._stopPressRepeat();
			this._repeatTimerInitialDelay = 0;
			this._repeatTimer = setInterval($.proxy(function() {
				this._repeatTimerInitialDelay += this.options.autorepeatInterval;
				if (this.options.autorepeatDelay < this._repeatTimerInitialDelay) {
					$.proxy(callback, this)();
				}
			},this), this.options.autorepeatInterval * 100);
		},

		_stopPressRepeat : function() {
			VLH.undefine(this,"_repeatTimer",clearInterval);
		},

		_handleSend : function(value)
		{
			if (this._canSend)
			{
				if (this._sliderChanged)
				{
					this._sliderChanged = false;
					VL.SetDeviceValue(this.options.devices[_outIndex].handle,value, $.proxy(function(){
						this._canSend = true;
						this._handleSend(this._sliderElement.slider("value"));
					},this));
					this._canSend = false;
					VL.InvokeEvent("USER_ACTIVITY");
				}
			}
		},

		_onStateChange : function() {
			if (VL.STATE == VL.CS.CONNECTED_CONTROL){
				this._changeState("enable");
			} else {
				this._changeState("disable");
			}
		},

		_changeState : function(state){
			this._minusButtonElement.button(state);
			this._sliderElement.slider(state);
			this._plusButtonElement.button(state);
		}
	});

	$.widget("my.VLConnection" , {

		options : {
			sound : '' //2DO
		},

		_create: function(){
			VL.RegisterEventListener("STATE_CHANGE",this._onStateChange,this);
			VL.RegisterEventListener("QUEUE_UPDATED",this._onStateChange,this);
			this.element.html(" &nbsp; ");
		},

		_onStateChange : function(){
			switch (VL.STATE)
			{
				case VL.CS.CONNECTED_CONTROL:
					if (VL.Timeleft>=0)
					{
						num = " "+Math.ceil(VL.Timeleft) + "&nbsp;"+VLH.L("SEC")+" ";
						if (VL.Timeleft>120)
						{
							num = " "+Math.ceil(VL.Timeleft/60) + "&nbsp;"+VLH.L("MIN")+" ";
						}

						this.element.html(VLH.L("CONTROL_TIME_LEFT_PREFIX") + num + VLH.L("CONTROL_TIME_LEFT_SUFFIX"));
					}
					else
					{
						this.element.html("");
					}
					break;

				case VL.CS.CONNECTED_WAIT:
					if (VL.Timeleft>=0)
					{
						num = " "+Math.ceil(VL.Timeleft) + "&nbsp;"+VLH.L("SEC")+" ";
						if (VL.Timeleft>120)
						{
							num = " "+Math.ceil(VL.Timeleft/60) + "&nbsp;"+VLH.L("MIN")+" ";
						}

						this.element.html(VLH.L("WAIT_TIME_LEFT_PREFIX") + num + VLH.L("WAIT_TIME_LEFT_SUFFIX"));
					}
					else
					{
						this.element.html("");
					}
					break;

				case VL.CS.ERROR:
					var message = (VL.ErrorMessage[0] == '#') ? VL.ErrorMessage.substr(1) : VLH.L(VL.ErrorMessage);
					this.element.html(VLH.L("ERROR")+": "+message);
					break;

				case VL.CS.INIT:
					this.element.html(VLH.L("INITIALIZING"));
					break;
			}
		}
	});

	$.widget("my.VLQueueDisplay" , {
		options : {
			refresh : 1000,
		},

		_create: function () {
			if (this.element.length != 1) {
				alert("Queue display error: Please apply widget to exactly one element!");
			} else {
				this._setList([]);
				this._update();
			}
		},

		_update : function() {
			VL.QueryQueueList($.proxy(function(result) {
				this._setList(result.Users);
				setTimeout($.proxy(this._update, this), this.options.refresh);
			}, this));
		},

		_setList : function(users) {
			var lines = "";
			for (var i = 0; i < users.length; ++i) {
				var row_class = ((i & 1) == 1) ? "queue_list_entry queue_list_entry_even" : " queue_list_entry queue_list_entry_odd";
				lines += "<tr><td class='" + row_class + "'>" + users[i].IP[0] + "." + users[i].IP[1] + "." + users[i].IP[2] + "." + users[i].IP[3] + "</td></tr>";
			}
			this.element.html("<table class='queue_frame'><tr><td><table class='queue_list'>" + lines + "</table></td></tr></table>");
		}
	});


}(jQuery));
